package com.grapeshot.halfnes;
//HalfNES, Copyright Andrew Hoffman, October 2010

import java.awt.image.*;
import java.util.Arrays;

public class PPU {

    public PPURAM ppuram;
    private int scanline = 0;
    private int oamaddr = 0;
    private int loopyV = 0x0;//ppu memory pointer
    private int loopyT = 0x0;//temp pointer
    private int loopyX = 0;//fine x scroll
    public int[] OAM = new int[256];
    private boolean even = true;
    private boolean bgpattern = true;
    private boolean sprpattern = false;
    private int readbuffer = 0;
    public int[] ppuregs = new int[0x8];
    private DebugUI debuggui;
    //no longer using the hash tables for tile caches because they actually SLOW it DOWN a bit now.
//    private HashMap<Integer, int[]> tilebuff = new HashMap<Integer, int[]>(
//            1024, 1);
    //private HashMap<Integer, int[]> debugbuff;
    private final static int TRANSPARENT = 0x00000000;
    // NES Color Palette (NTSC) not including emphasis bits
    // TODO load this from a file instead
    //uses the function to premultiply with alpha channel
    private final static int[] nescolor = GetNESColors();
    private int vraminc = 1;
    private int[] bitmap = new int[240 * 256];
    private final static boolean PPUDEBUG = false;
    private BufferedImage newBuff;

    public PPU(final PPURAM ppuram) {
        this.ppuram = ppuram;
        Arrays.fill(OAM, 0xff);
        Arrays.fill(ppuregs, 0x00);
        if (PPUDEBUG) {
            newBuff = new BufferedImage(512, 480, BufferedImage.TYPE_3BYTE_BGR);
            debuggui = new DebugUI();
            debuggui.run();
            //debugbuff = new HashMap<Integer, int[]>(1024, 1);
        }
    }

    public int read(final int regnum) {
        switch (regnum) {
            case 2:
                even = true;
                final int tmp = ppuregs[2];
                ppuregs[2] &= 0x7f;//turn off vblank flag
                return tmp;
            case 4:
                // reading this is NOT reliable but some games do it anyways
                // TODO figure it out to fix Micro Machines
                return OAM[oamaddr];
            case 7:
                // PPUDATA
                // correct behavior. read is delayed by one
                // -unless- is a read from sprite pallettes
                if ((loopyV & 0x3fff) < 0x3f00) {
                    final int temp = readbuffer;
                    readbuffer = ppuram.read(loopyV & 0x3fff);
                    loopyV += vraminc;
                    return temp;
                } else {
                    readbuffer = ppuram.read((loopyV & 0x3fff) - 0x1000);
                    final int temp = ppuram.read(loopyV);
                    loopyV += vraminc;
                    return temp;
                }

            // and don't increment on read
            default:
                return 0x20; // open bus
        }
    }

    public void write(final int regnum, final int data) {
        switch (regnum) {
            case 0:
                ppuregs[0] = (short) data;
                vraminc = (utils.getbit(data, 2) ? 32 : 1);
                //set 2 bits of vram address (nametable select)
                loopyT &= ~0xc00;
                loopyT += (data & 3) << 10;
                break;
            case 1:
                ppuregs[1] = (short) data;
            case 3:
                // PPUOAMADDR (2003)
                // most games just write zero and use the dma
                oamaddr = data & 0xff;
                break;
            case 4:
                // PPUOAMDATA(2004)
                OAM[oamaddr & 0xff] = data;
                oamaddr++;
                oamaddr &= 0xff;
                // games don't write this directly anyway
                break;

            // PPUSCROLL(2005)
            case 5:
                if (even) {
                    // horizontal scroll
                    loopyT &= ~0x1f;
                    loopyX = data & 7;
                    loopyT += data >> 3;

                    even = false;
                } else {
                    // vertical scroll
                    loopyT &= ~0x7000;
                    loopyT += ((data & 7) << 12);
                    loopyT &= ~0x3e0;
                    loopyT += (data >> 3) << 5;
                    even = true;

                }
                break;

            case 6:
                // PPUADDR (2006)
                if (even) {
                    // high byte
                    loopyT &= 0xc0ff;
                    loopyT += ((data & 0x3f) << 8);
                    loopyT &= 0x3fff;
                    even = false;
                } else {
                    loopyT &= 0xff00;
                    loopyT += data;
                    loopyV = loopyT;
                    even = true;

                }
                break;
            case 7:
                // PPUDATA
                if (renderingisoff()) {
                    // if rendering is off its safe to write
                    ppuram.write((loopyV & 0x3fff), data);
                    loopyV += vraminc;
                } else {
                    //System.err.println("dropped write");
                    // write anyway though until i figure out which wrong thing to do
                    //also since the ppu doesn't get pixel level timing right now,
                    //i can't detect hblank.
                    ppuram.write((loopyV & 0x3fff), data);
                    loopyV += vraminc;
                }
            // increments on write but NOT on read
            default:
                break;
        }
    }

    public boolean renderingisoff() {
        // tells when it's ok to write to the ppu
        return (scanline >= 240) //|| (pixel > 256)
                || (!utils.getbit(ppuregs[1], 3));
    }

    public boolean mmc3CounterClocking() {
        return (bgpattern != sprpattern) && !renderingisoff();
    }

    public void drawLine(final int scanline) {
        //this contains probably more magic numbers than the rest of the program combined.
        //TODO: define some static bitmasks to manipulate the address through, instead
        bgpattern = utils.getbit(ppuregs[0], 4);
        sprpattern = utils.getbit(ppuregs[0], 3);
        int bgcolor;
        final int bufferoffset = scanline * 256;
        //TODO: Simplify Logic
        bgcolor = nescolor[ppuram.pal[0] % nescolor.length];
        if (utils.getbit(ppuregs[1], 3)) {
            //System.err.println("BG on @ "+ scanline);
            // if bg is on, draw tiles.
            if (scanline == 0) {
                //update whole scroll
                loopyV = loopyT;
            } else {
                //update hscroll only
                loopyV &= ~0x41f;
                loopyV |= loopyT & 0x41f;
            }
            //draw background
            int ntoffset = (loopyV & 0xc00) | 0x2000;
            int attroffset = (loopyV & 0xc00) | 0x2000 + 0x3c0;
            boolean horizWrap = false;
            for (int tilenum = 0; tilenum < 33; ++tilenum) {
                //for each tile in row
                if ((tilenum * 8 + (((loopyV & 0x1f) << 3) + loopyX)) > 255 && !horizWrap) {
                    //if scrolling off the side of the nametable, bump address to next nametable
                    ntoffset ^= 0x400;
                    ntoffset -= 32;
                    attroffset ^= 0x400;
                    horizWrap = true;
                }
                //get palette number from attribute table byte
                final int palettenum = getattrtbl(attroffset, (ntoffset + loopyV + tilenum) & 0x1f, (((ntoffset + loopyV + tilenum) & 0x3e0) >> 5));
                final int tileaddr = ppuram.read(ntoffset + (loopyV & 0x3ff) + tilenum) * 16 + (bgpattern ? 0x1000 : 0);
                final int[] tile = getTile(tileaddr, palettenum * 4, (loopyV & 0x7000) >> 12, false, false);
                //now put inna buffer
                final int xpos = tilenum * 8 - loopyX;//not quite right yet
                for (int pxl = 0; pxl < 8; ++pxl) {
                    if ((pxl + xpos) < 256 && (pxl + xpos) >= 0) { //it's not off the screen
                        bitmap[pxl + xpos + bufferoffset] = tile[pxl];
                    }
                }
            }
            //increment loopy_v to next row of tiles
            int newfinescroll = loopyV & 0x7000;
            newfinescroll += 0x1000;
            loopyV &= ~0x7000;
            if (newfinescroll > 0x7000) {
                //reset the fine scroll bits and increment tile address to next row
                loopyV += 32;
            } else {
                //increment the fine scroll
                loopyV += newfinescroll;
            }
            if (((loopyV >> 5) & 0x1f) >= 30) {
                //if incrementing loopy_v to the next row pushes us into the next
                //nametable, zero the "row" bits and go to next nametable
                loopyV &= ~0x3e0;
                loopyV ^= 0x800;
                ntoffset += 0x440;
                attroffset += 0x7c0;
            }
        } else {
            //System.err.println("BG off @ "+ scanline);
            //if rendering is off draw either the background color OR
            //if the PPU address points to the palette, draw that color instead.
            bgcolor = nescolor[((loopyV > 0x3f00 && loopyV < 0x3fff) ? ppuram.read(loopyV) : ppuram.pal[0]) % nescolor.length];
            Arrays.fill(bitmap, bufferoffset, bufferoffset + 256, bgcolor);
        }
        //draw sprites on top of whatever we had
        drawSprites(scanline, bgcolor);
        //hide leftmost 8 pixels if that flag is on
        if (!utils.getbit(ppuregs[1], 1)) {
            for (int i = 0; i < 8; ++i) {
                bitmap[i + bufferoffset] = bgcolor;
            }
        }
    }

    private void drawSprites(final int scanline, final int bgcolor) {
        if (!utils.getbit(ppuregs[1], 4)) {
            return; //return immediately if sprites are disabled
        }
        final int bufferoffset = 256 * scanline;
        bgpattern = utils.getbit(ppuregs[0], 4);
        sprpattern = utils.getbit(ppuregs[0], 3);
        int ypos, xpos, offset, tilefetched;
        int found = 0;
        final boolean spritesize = utils.getbit(ppuregs[0], 5);
        int[] tile;
        int[] spritebuffer = new int[256];
        boolean[] bgmask = new boolean[256];
        Arrays.fill(spritebuffer, TRANSPARENT);
        for (int i = 63; i >= 0; --i) {
            //for each sprite, first we cull the non-visible ones
            ypos = OAM[(4 * i)] + 1;
            offset = scanline - ypos;
            if (ypos > scanline || offset > (spritesize ? 15 : 7)) {
                //sprite is out of range vertically
                continue;
            }
            //if we're here it's a valid renderable sprite
            ++found;
            if (found > 8) {
                //if more than 8 sprites, set overflow bit (but render anyway)
                ppuregs[2] |= 0x20;
            }
            final int oamextra = OAM[(4 * i) + 2];
            boolean background = utils.getbit(oamextra, 5);
            xpos = OAM[(4 * i) + 3];
            final int sprpal = ((oamextra & 3) + 4) * 4;
            if (utils.getbit(oamextra, 7)) {
                //if sprite is flipped vertically, reverse the offset
                offset = (spritesize ? 15 : 7) - offset;
            }
            //now correction for the fact that 8x16 tiles are 2 separate tiles
            if (offset > 7) {
                offset += 8;
            }
            //get tile address (8x16 sprites can use both pattern tbl pages but only the even tiles)
            final int tilenum = OAM[(4 * i) + 1];
            if (spritesize) {
                tilefetched = (utils.getbitI(tilenum, 0) * 0x1000)
                        + (tilenum & 0xfe) * 16;
            } else {
                tilefetched = tilenum * 16
                        + ((sprpattern) ? 0x1000 : 0);
            }
            final boolean hflip = utils.getbit(oamextra, 6);
            tile = getTile(tilefetched, sprpal, offset, true, hflip);
            //now we render that sprite on the sprite buffer
            for (int pxl = 0; pxl < 8; ++pxl) {
                if (pxl + xpos < 256 //it's not off the screen
                        && tile[pxl] != TRANSPARENT) { //and that pixel of tile is opaque
                    spritebuffer[pxl + xpos] = tile[pxl];
                    bgmask[pxl + xpos] = background; //save whether it was a background sprite or not for next
                }
            }
        }
        //now render that completed line of sprite pixels over the background
        for (int i = 0; i < 256; ++i) {
            //for each pixel in the line
            //if there's a sprite pixel there
            if (spritebuffer[i] == TRANSPARENT) {
                continue;
            }
            //and, if it's a BACKGROUND sprite pixel, it's being drawn over the background color
            //(because if it was being drawn over a foreground color then it wouldn't be in the background, now would it)
            if (bgmask[i] && (bitmap[i + bufferoffset] != bgcolor)) {
                continue;
            }
            //then copy that to the line buffer
            bitmap[i + bufferoffset] = spritebuffer[i];
        }
    }

    private int getattrtbl(final int ntstart, final int tilex, final int tiley) {
        final int base = ntstart + (tilex >> 2) + 8 * (tiley >> 2);
        if ((tilex & 2) == 0) {
            if ((tiley & 2) == 0) {
                return ppuram.read(base) & 3;
            } else {
                return (ppuram.read(base) >> 4) & 3;
            }
        } else {
            if ((tiley & 2) == 0) {
                return (ppuram.read(base) >> 2) & 3;
            } else {
                return (ppuram.read(base) >> 6) & 3;
            }
        }
    }

    public void debugdraw() {
        //old code, left for dumping out VRAM to debug window.
        //SLOW.
        final boolean tilemode = false;
        if (tilemode) {
            for (int i = 0; i < 32; ++i) {
                for (int j = 0; j < 30; ++j) {
                    newBuff.setRGB(i * 8, j * 8, 8, 8, oldgettile(ppuram.read(0x2000 + i + 32 * j) * 16 + (bgpattern ? 0x1000 : 0)), 0, 8);
                }
            }
            for (int i = 0; i < 32; ++i) {
                for (int j = 0; j < 30; ++j) {
                    newBuff.setRGB(i * 8 + 255, j * 8, 8, 8, oldgettile(ppuram.read(0x2400 + i + 32 * j) * 16 + (bgpattern ? 0x1000 : 0)), 0, 8);
                }
            }
            for (int i = 0; i < 32; ++i) {
                for (int j = 0; j < 30; ++j) {
                    newBuff.setRGB(i * 8, j * 8 + 239, 8, 8, oldgettile(ppuram.read(0x2800 + i + 32 * j) * 16 + (bgpattern ? 0x1000 : 0)), 0, 8);
                }
            }
            for (int i = 0; i < 32; ++i) {
                for (int j = 0; j < 30; ++j) {
                    newBuff.setRGB(i * 8 + 255, j * 8 + 239, 8, 8, oldgettile(ppuram.read(0x2C00 + i + 32 * j) * 16 + (bgpattern ? 0x1000 : 0)), 0, 8);
                }
            }
        } else {
            //draw the tileset instead
            for (int i = 0; i < 16; ++i) {
                for (int j = 0; j < 32; ++j) {
                    newBuff.setRGB(i * 8, j * 8, 8, 8, oldgettile((i + 16 * j) * 16), 0, 8);
                }
            }
            //and the palettes.
            for (int i = 0; i < 32; ++i) {
                for (int j = 0; j < 16; ++j) {
                    for (int k = 0; k < 16; ++k) {
                        newBuff.setRGB(j + i * 16, k + 256, nescolor[ppuram.pal[i]]);
                    }
                }
            }
        }
        debuggui.setFrame(newBuff);
        //debugbuff.clear();
    }

    public int[] oldgettile(final int patterntblptr) {
        // this'll be really really slow
        //for debug only
        int[] dat = new int[64];
        for (int i = 0; i < 8; ++i) {
            //per line of tile ( 1 byte)
            for (int j = 0; j < 8; ++j) {
                //per pixel(1 bit)
                dat[8 * i + j] = ((utils.getbit(ppuram.read(i + patterntblptr), 7 - j)) ? 0x555555 : 0) + ((utils.getbit(ppuram.read(i + patterntblptr + 8), 7 - j)) ? 0xaaaaaa : 0);
            }
        }
        return dat;
    }

    public int spriteZeroFirstLine() {
        //finds first filled line of sprite zero for timing purposes
        //TODO: fix the spurious PPU reads this causes if I ever do an accurate
        //mmc3 scanline counters
        if (!utils.getbit(ppuregs[0], 5)) {
            //8x8 sprites
            final int sprite0loc = OAM[1] * 16 + (sprpattern ? 0x1000 : 0);
            for (int j = 0; j < 8; ++j) {
                if (((ppuram.read(j + sprite0loc)) != 0) && ((ppuram.read(j + sprite0loc + 8)) != 0)) {
                    return j;
                }
            }
            return 7;
        } else {
            //8x16 sprites
            int sprite0loc = (OAM[1] >> 1) * 32 + (utils.getbitI(OAM[1], 0) * 0x1000);
            for (int j = 0; j < 8; ++j) {
                if (((ppuram.read(j + sprite0loc)) != 0) || ((ppuram.read(j + sprite0loc + 8)) != 0)) {
                    return j;
                }
            }
            sprite0loc += 16;
            for (int j = 0; j < 8; ++j) {
                if (((ppuram.read(j + sprite0loc)) != 0) || ((ppuram.read(j + sprite0loc + 8)) != 0)) {
                    return j + 8;
                }
            }
            return 15;
        }
    }

    public BufferedImage renderFrame() {
        if (PPUDEBUG) {
            debugdraw();
        }
        return getImageFromArray(bitmap, 256, 240);
    }

    public static BufferedImage getImageFromArray(final int[] bitmap, final int width, final int height) {
        final BufferedImage image = new BufferedImage(width, height, BufferedImage.TYPE_INT_ARGB_PRE);
        final WritableRaster raster = image.getRaster();
        final int[] pixels = ((DataBufferInt) raster.getDataBuffer()).getData();
        System.arraycopy(bitmap, 0, pixels, 0, bitmap.length);
        return image;
    }

    public int[] getTile(final int tileptr, final int paletteindex, final int off, final boolean isSprite, final boolean hflip) {
        //returns an 8 pixel line of tile data fron given PPU ram location
        //with given offset and given palette.
//        int hashkey = ((tileptr + off) << 5) + paletteindex + (hflip ? 1 : 0);
//        //I really didn't want to keep the hash table since it's so troublesome.
//        //also, now the overhead of finding the index is same as for rendering the line of the sprite.
//        if(tilebuff.containsKey(hashkey)){
//            return tilebuff.get(hashkey);
//        }
        int[] data = new int[8];
        final int[] color = {
            isSprite ? TRANSPARENT : nescolor[ppuram.pal[0]],
            nescolor[ppuram.pal[paletteindex + 1]],
            nescolor[ppuram.pal[paletteindex + 2]],
            nescolor[ppuram.pal[paletteindex + 3]]
        };
        // per line of tile ( 1 byte)
        int linelowbits = ppuram.read(off + tileptr);
        int linehighbits = ppuram.read(off + tileptr + 8);
        if (hflip) {
            for (int j = 0; j < 8; ++j) {
                // per pixel(1 bit)
                data[j] = color[((linehighbits & 1) << 1) + (linelowbits & 1)];
                linehighbits >>= 1;
                linelowbits >>= 1;
            }
        } else {
            for (int j = 0; j < 8; ++j) {
                // per pixel(1 bit)
                data[7 - j] = color[((linehighbits & 1) << 1) + (linelowbits & 1)];
                linehighbits >>= 1;
                linelowbits >>= 1;
            }
        }
//        tilebuff.put(hashkey, data);
        return data;
    }

    public void clearTileCache() {
        //tilebuff.clear();
    }

    private static int[] GetNESColors() {
        //just or's all the colors with opaque alpha, to save time in other loops
        int[] colorarray = {0x757575, 0x271B8F, 0x0000AB,
            0x47009F, 0x8F0077, 0xAB0013, 0xA70000, 0x7F0B00, 0x432F00,
            0x004700, 0x005100, 0x003F17, 0x1B3F5F, 0x000000, 0x000000,
            0x000000, 0xBCBCBC, 0x0073EF, 0x233BEF, 0x8300F3, 0xBF00BF,
            0xE7005B, 0xDB2B00, 0xCB4F0F, 0x8B7300, 0x009700, 0x00AB00,
            0x00933B, 0x00838B, 0x000000, 0x000000, 0x000000, 0xFFFFFF,
            0x3FBFFF, 0x5F97FF, 0xA78BFD, 0xF77BFF, 0xFF77B7, 0xFF7763,
            0xFF9B3B, 0xF3BF3F, 0x83D313, 0x4FDF4B, 0x58F898, 0x00EBDB,
            0x000000, 0x000000, 0x000000, 0xFFFFFF, 0xABE7FF, 0xC7D7FF,
            0xD7CBFF, 0xFFC7FF, 0xFFC7DB, 0xFFBFB3, 0xFFDBAB, 0xFFE7A3,
            0xE3FFA3, 0xABF3BF, 0xB3FFCF, 0x9FFFF3, 0x000000, 0x000000,
            0x000000};
        for (int i = 0; i < colorarray.length; ++i) {
            colorarray[i] |= 0xff000000;
        }
        return colorarray;
    }
}
